@@ -19,7 +19,7 @@
 
 #include "link.h"
 
-#include "akonadiconnection.h"
+#include "connection.h"
 #include "handlerhelper.h"
 #include "storage/datastore.h"
 #include "storage/itemqueryhelper.h"
@@ -30,7 +30,7 @@
 #include "imapstreamparser.h"
 #include <storage/collectionqueryhelper.h>
 
-using namespace Akonadi;
+using namespace Akonadi::Server;
 
 Link::Link( Scope::SelectionScope scope, bool create )
   : Handler()
