@@ -36,6 +36,8 @@
 #include "handler.h"
 #include "collectionqueryhelper.h"
 #include "akonadischema.h"
+#include "parttypehelper.h"
+#include "querycache.h"
 
 #include <QtCore/QCoreApplication>
 #include <QtCore/QDir>
@@ -54,7 +56,7 @@
 #include <QtSql/QSqlField>
 #include <QtSql/QSqlQuery>
 
-using namespace Akonadi;
+using namespace Akonadi::Server;
 
 static QMutex sTransactionMutex;
 bool DataStore::s_hasForeignKeyConstraints = false;
@@ -65,16 +67,15 @@ bool DataStore::s_hasForeignKeyConstraints = false;
 /***************************************************************************
  *   DataStore                                                           *
  ***************************************************************************/
-DataStore::DataStore() :
-  QObject(),
-  m_dbOpened( false ),
-  m_transactionLevel( 0 ),
-  mNotificationCollector( new NotificationCollector( this ) ),
-  m_keepAliveTimer( 0 )
+DataStore::DataStore()
+  : QObject()
+  , m_dbOpened( false )
+  , m_transactionLevel( 0 )
+  , mNotificationCollector( new NotificationCollector( this ) )
+  , m_keepAliveTimer( 0 )
 {
   open();
 
-  m_transactionLevel = 0;
   NotificationManager::self()->connectNotificationCollector( mNotificationCollector );
 
   if ( DbConfig::configuredDatabase()->driverName() == QLatin1String( "QMYSQL" ) ) {
@@ -117,7 +118,7 @@ void DataStore::open()
   DbConfig::configuredDatabase()->initSession( m_database );
 }
 
-void Akonadi::DataStore::close()
+void DataStore::close()
 {
 
   if ( m_keepAliveTimer ) {
@@ -135,6 +136,7 @@ void Akonadi::DataStore::close()
     rollbackTransaction();
   }
 
+  QueryCache::clear();
   m_database.close();
   m_database = QSqlDatabase();
   QSqlDatabase::removeDatabase( m_connectionName );
@@ -142,13 +144,13 @@ void Akonadi::DataStore::close()
   m_dbOpened = false;
 }
 
-bool Akonadi::DataStore::init()
+bool DataStore::init()
 {
   Q_ASSERT( QThread::currentThread() == QCoreApplication::instance()->thread() );
 
   AkonadiSchema schema;
   DbInitializer::Ptr initializer = DbInitializer::createInstance( m_database, &schema );
-  if (! initializer->run() ) {
+  if ( !initializer->run() ) {
     akError() << initializer->errorMsg();
     return false;
   }
@@ -159,6 +161,11 @@ bool Akonadi::DataStore::init()
     return false;
   }
 
+  if ( !initializer->updateIndexesAndConstraints() ) {
+    akError() << initializer->errorMsg();
+    return false;
+  }
+
   // enable caching for some tables
   MimeType::enableCache( true );
   Flag::enableCache( true );
@@ -168,9 +175,9 @@ bool Akonadi::DataStore::init()
   return true;
 }
 
-QThreadStorage<DataStore*> instances;
+QThreadStorage<DataStore *> instances;
 
-DataStore * Akonadi::DataStore::self()
+DataStore *DataStore::self()
 {
   if ( !instances.hasLocalData() ) {
     instances.setLocalData( new DataStore() );
@@ -182,62 +189,62 @@ DataStore * Akonadi::DataStore::self()
 
 bool DataStore::setItemsFlags( const PimItem::List &items, const QVector<Flag> &flags )
 {
-  // first delete all old flags of this pim item
   QSet<QByteArray> removedFlags;
   QSet<QByteArray> addedFlags;
-  QVariantList ids;
   QVariantList insIds;
   QVariantList insFlags;
+  Query::Condition delConds( Query::Or );
 
   Q_FOREACH ( const PimItem &item, items ) {
     Q_FOREACH ( const Flag &flag, item.flags() ) {
-      if ( !removedFlags.contains( flag.name().toLatin1() ) ) {
+      if ( !flags.contains( flag ) ) {
         removedFlags << flag.name().toLatin1();
+        Query::Condition cond;
+        cond.addValueCondition( PimItemFlagRelation::leftFullColumnName(), Query::Equals, item.id() );
+        cond.addValueCondition( PimItemFlagRelation::rightFullColumnName(), Query::Equals, flag.id() );
+        delConds.addCondition(cond);
       }
     }
 
-    // create a bind values for the insert query - every item repeats exactly
-    // flags.count()-times
-    for ( int i = 0; i < flags.count(); ++i ) {
-      insIds << item.id();
+    Q_FOREACH ( const Flag &flag, flags ) {
+      if ( !item.flags().contains( flag ) ) {
+        addedFlags << flag.name().toLatin1();
+        insIds << item.id();
+        insFlags << flag.id();
+      }
     }
-
-    ids << item.id();
   }
 
-  // Clear all flags of all given items at once
-  QueryBuilder qb( PimItemFlagRelation::tableName(), QueryBuilder::Delete );
-  Query::Condition cond;
-  cond.addValueCondition( PimItemFlagRelation::leftFullColumnName(), Query::In, ids );
-  qb.addCondition( cond );
-  if ( !qb.exec() ) {
-    return false;
-  }
-
-  // create bind values for the insert quest - every flags repeats exactly
-  // items.count()-times
-  Q_FOREACH ( const Flag &flag, flags ) {
-    for ( int i = 0; i < items.count(); ++i ) {
-      insFlags << flag.id();
+  if ( !removedFlags.empty() ) {
+    QueryBuilder qb( PimItemFlagRelation::tableName(), QueryBuilder::Delete );
+    qb.addCondition( delConds );
+    if ( !qb.exec() ) {
+      return false;
     }
-
-    addedFlags << flag.name().toLatin1();
   }
 
-  QueryBuilder qb2( PimItemFlagRelation::tableName(), QueryBuilder::Insert );
-  qb2.setColumnValue( PimItemFlagRelation::leftColumn(), insIds );
-  qb2.setColumnValue( PimItemFlagRelation::rightColumn(), insFlags );
-  qb2.setIdentificationColumn( QString() );
-  if ( !qb2.exec() ) {
-    return false;
+  if ( !addedFlags.empty() ) {
+    QueryBuilder qb2( PimItemFlagRelation::tableName(), QueryBuilder::Insert );
+    qb2.setColumnValue( PimItemFlagRelation::leftColumn(), insIds );
+    qb2.setColumnValue( PimItemFlagRelation::rightColumn(), insFlags );
+    qb2.setIdentificationColumn( QString() );
+    if ( !qb2.exec() ) {
+      return false;
+    }
+  }
+
+  if ( addedFlags.empty() && removedFlags.empty() ) {
+    // no changes done, notification not needed
+    return true;
   }
 
   mNotificationCollector->itemsFlagsChanged( items, addedFlags, removedFlags );
+
   return true;
 }
 
 bool DataStore::doAppendItemsFlag( const PimItem::List &items, const Flag &flag,
-                                   const QSet<Entity::Id> &existing, const Collection &col)
+                                   const QSet<Entity::Id> &existing, const Collection &col )
 {
   QVariantList flagIds;
   QVariantList appendIds;
@@ -272,7 +279,7 @@ bool DataStore::doAppendItemsFlag( const PimItem::List &items, const Flag &flag,
 }
 
 bool DataStore::appendItemsFlags( const PimItem::List &items, const QVector<Flag> &flags,
-                                  bool& flagsChanged, bool checkIfExists,
+                                  bool &flagsChanged, bool checkIfExists,
                                   const Collection &col )
 {
   QSet<QByteArray> added;
@@ -326,9 +333,9 @@ bool DataStore::removeItemsFlags( const PimItem::List &items, const QVector<Flag
   Q_FOREACH ( const PimItem &item, items ) {
     itemsIds << item.id();
     for ( int i = 0; i < flags.count(); ++i ) {
-      const QByteArray flagName = flags[ i ].name().toLatin1();
-      if ( !flagsIds.contains( flagName ) ) {
-        flagsIds << flags[ i ].id();
+      const QByteArray flagName = flags[i].name().toLatin1();
+      if ( !removedFlags.contains( flagName ) ) {
+        flagsIds << flags[i].id();
         removedFlags << flagName;
       }
     }
@@ -348,16 +355,191 @@ bool DataStore::removeItemsFlags( const PimItem::List &items, const QVector<Flag
   return true;
 }
 
+/* --- ItemTags ----------------------------------------------------- */
+
+bool DataStore::setItemsTags( const PimItem::List &items, const Tag::List &tags )
+{
+  QSet<qint64> removedTags;
+  QSet<qint64> addedTags;
+  QVariantList insIds;
+  QVariantList insTags;
+  Query::Condition delConds( Query::Or );
+
+  Q_FOREACH ( const PimItem &item, items ) {
+    Q_FOREACH ( const Tag &tag, item.tags() ) {
+      if ( !tags.contains( tag ) ) {
+        // Remove tags from items that had it set
+        removedTags << tag.id();
+        Query::Condition cond;
+        cond.addValueCondition( PimItemTagRelation::leftFullColumnName(), Query::Equals, item.id() );
+        cond.addValueCondition( PimItemTagRelation::rightFullColumnName(), Query::Equals, tag.id() );
+        delConds.addCondition(cond);
+      }
+    }
+
+    Q_FOREACH ( const Tag &tag, tags ) {
+      if ( !item.tags().contains( tag ) ) {
+        // Add tags to items that did not have the tag
+        addedTags << tag.id();
+        insIds << item.id();
+        insTags << tag.id();
+      }
+    }
+  }
+
+  if ( !removedTags.empty() ) {
+    QueryBuilder qb( PimItemTagRelation::tableName(), QueryBuilder::Delete );
+    qb.addCondition( delConds );
+    if ( !qb.exec() ) {
+      return false;
+    }
+  }
+
+  if ( !addedTags.empty() ) {
+    QueryBuilder qb2( PimItemTagRelation::tableName(), QueryBuilder::Insert );
+    qb2.setColumnValue( PimItemTagRelation::leftColumn(), insIds );
+    qb2.setColumnValue( PimItemTagRelation::rightColumn(), insTags );
+    qb2.setIdentificationColumn( QString() );
+    if ( !qb2.exec() ) {
+      return false;
+    }
+  }
+
+  if ( addedTags.empty() && removedTags.empty() ) {
+    // no changes done, notification not needed
+    return true;
+  }
+
+  mNotificationCollector->itemsTagsChanged( items, addedTags, removedTags );
+
+  return true;
+}
+
+bool DataStore::doAppendItemsTag( const PimItem::List &items, const Tag &tag,
+                                   const QSet<Entity::Id> &existing, const Collection &col )
+{
+  QVariantList tagIds;
+  QVariantList appendIds;
+  PimItem::List appendItems;
+  Q_FOREACH ( const PimItem &item, items ) {
+    if ( existing.contains( item.id() ) ) {
+      continue;
+    }
+
+    tagIds << tag.id();
+    appendIds << item.id();
+    appendItems << item;
+  }
+
+  if ( appendItems.isEmpty() ) {
+    return true; // all items have the desired tags already
+  }
+
+  QueryBuilder qb2( PimItemTagRelation::tableName(), QueryBuilder::Insert );
+  qb2.setColumnValue( PimItemTagRelation::leftColumn(), appendIds );
+  qb2.setColumnValue( PimItemTagRelation::rightColumn(), tagIds );
+  qb2.setIdentificationColumn( QString() );
+  if ( !qb2.exec() ) {
+    akDebug() << "Failed to execute query:" << qb2.query().lastError();
+    return false;
+  }
+
+  mNotificationCollector->itemsTagsChanged( appendItems, QSet<qint64>() << tag.id(),
+                                             QSet<qint64>(), col );
+
+  return true;
+}
+
+bool DataStore::appendItemsTags( const PimItem::List &items, const Tag::List &tags,
+                                  bool &tagsChanged, bool checkIfExists,
+                                  const Collection &col )
+{
+  QSet<QByteArray> added;
+
+  QVariantList itemsIds;
+  Q_FOREACH ( const PimItem &item, items ) {
+    itemsIds.append( item.id() );
+  }
+
+  tagsChanged = false;
+  Q_FOREACH ( const Tag &tag, tags ) {
+    QSet<PimItem::Id> existing;
+    if ( checkIfExists ) {
+      QueryBuilder qb( PimItemTagRelation::tableName(), QueryBuilder::Select );
+      Query::Condition cond;
+      cond.addValueCondition( PimItemTagRelation::rightColumn(), Query::Equals, tag.id() );
+      cond.addValueCondition( PimItemTagRelation::leftColumn(), Query::In, itemsIds );
+      qb.addColumn( PimItemTagRelation::leftColumn() );
+      qb.addCondition( cond );
+
+      if ( !qb.exec() ) {
+        akDebug() << "Failed to execute query:" << qb.query().lastError();
+        return false;
+      }
+
+      QSqlQuery query = qb.query();
+      if ( query.size() == items.count() ) {
+        continue;
+      }
+
+      tagsChanged = true;
+
+      while ( query.next() ) {
+        existing << query.value( 0 ).value<PimItem::Id>();
+      } }
+
+    if ( !doAppendItemsTag( items, tag, existing, col ) ) {
+      return false;
+    }
+  }
+
+  return true;
+}
+
+bool DataStore::removeItemsTags( const PimItem::List &items, const Tag::List &tags )
+{
+  QSet<qint64> removedTags;
+  QVariantList itemsIds;
+  QVariantList tagsIds;
+  Q_FOREACH ( const PimItem &item, items ) {
+    itemsIds << item.id();
+    for ( int i = 0; i < tags.count(); ++i ) {
+      const qint64 tagId = tags[i].id();
+      if ( !removedTags.contains( tagId ) ) {
+        tagsIds << tagId;
+        removedTags << tagId;
+      }
+    }
+  }
+
+  // Delete all given tags from all given items in one go
+  QueryBuilder qb( PimItemTagRelation::tableName(), QueryBuilder::Delete );
+  Query::Condition cond( Query::And );
+  cond.addValueCondition( PimItemTagRelation::rightFullColumnName(), Query::In, tagsIds );
+  cond.addValueCondition( PimItemTagRelation::leftFullColumnName(), Query::In, itemsIds );
+  qb.addCondition( cond );
+  if ( !qb.exec() ) {
+    return false;
+  }
+
+  mNotificationCollector->itemsTagsChanged( items, QSet<qint64>(), removedTags );
+  return true;
+}
+
 /* --- ItemParts ----------------------------------------------------- */
 
 bool DataStore::removeItemParts( const PimItem &item, const QList<QByteArray> &parts )
 {
-  Part::List existingParts = item.parts();
+  SelectQueryBuilder<Part> qb;
+  qb.addJoin( QueryBuilder::InnerJoin, PartType::tableName(), Part::partTypeIdFullColumnName(), PartType::idFullColumnName() );
+  qb.addValueCondition( Part::pimItemIdFullColumnName(), Query::Equals, item.id() );
+  qb.addCondition( PartTypeHelper::conditionFromFqNames( parts ) );
+
+  qb.exec();
+  Part::List existingParts = qb.result();
   Q_FOREACH ( Part part, existingParts ) {
-    if ( parts.contains( part.name().toLatin1() ) ) {
-      if ( !PartHelper::remove(&part) ) {
-        return false;
-      }
+    if ( !PartHelper::remove( &part ) ) {
+      return false;
     }
   }
 
@@ -367,12 +549,13 @@ bool DataStore::removeItemParts( const PimItem &item, const QList<QByteArray> &p
 
 bool DataStore::invalidateItemCache( const PimItem &item )
 {
-  // find all expired item parts
+  // find all payload item parts
   SelectQueryBuilder<Part> qb;
   qb.addJoin( QueryBuilder::InnerJoin, PimItem::tableName(), PimItem::idFullColumnName(), Part::pimItemIdFullColumnName() );
+  qb.addJoin( QueryBuilder::InnerJoin, PartType::tableName(), Part::partTypeIdFullColumnName(), PartType::idFullColumnName() );
   qb.addValueCondition( Part::pimItemIdFullColumnName(), Query::Equals, item.id() );
   qb.addValueCondition( Part::dataFullColumnName(), Query::IsNot, QVariant() );
-  qb.addValueCondition( QString::fromLatin1( "substr( %1, 1, 4 )" ).arg( Part::nameFullColumnName() ), Query::Equals, QLatin1String( "PLD:" ) );
+  qb.addValueCondition( PartType::nsFullColumnName(), Query::Equals, QLatin1String( "PLD" ) );
   qb.addValueCondition( PimItem::dirtyFullColumnName(), Query::Equals, false );
 
   if ( !qb.exec() ) {
@@ -403,7 +586,7 @@ bool DataStore::appendCollection( Collection &collection )
   return true;
 }
 
-bool Akonadi::DataStore::cleanupCollection(Collection &collection)
+bool DataStore::cleanupCollection( Collection &collection )
 {
   if ( !s_hasForeignKeyConstraints ) {
     return cleanupCollection_slow( collection );
@@ -433,8 +616,9 @@ bool Akonadi::DataStore::cleanupCollection(Collection &collection)
   }
 
   try {
-    while ( qb.query().next() )
+    while ( qb.query().next() ) {
       PartHelper::removeFile( PartHelper::resolveAbsolutePath( qb.query().value( 0 ).value<QByteArray>() ) );
+    }
   } catch ( const PartHelperException &e ) {
     akDebug() << e.what();
     return false;
@@ -445,7 +629,7 @@ bool Akonadi::DataStore::cleanupCollection(Collection &collection)
   return collection.remove();
 }
 
-bool DataStore::cleanupCollection_slow(Collection& collection)
+bool DataStore::cleanupCollection_slow( Collection &collection )
 {
   Q_ASSERT( !s_hasForeignKeyConstraints );
 
@@ -487,7 +671,7 @@ bool DataStore::cleanupCollection_slow(Collection& collection)
   return collection.remove();
 }
 
-static bool recursiveSetResourceId( const Collection & collection, qint64 resourceId )
+static bool recursiveSetResourceId( const Collection &collection, qint64 resourceId )
 {
   Transaction transaction( DataStore::self() );
 
@@ -524,7 +708,7 @@ static bool recursiveSetResourceId( const Collection & collection, qint64 resour
   return true;
 }
 
-bool Akonadi::DataStore::moveCollection( Collection & collection, const Collection &newParent )
+bool DataStore::moveCollection( Collection &collection, const Collection &newParent )
 {
   if ( collection.parentId() == newParent.id() ) {
     return true;
@@ -541,7 +725,7 @@ bool Akonadi::DataStore::moveCollection( Collection & collection, const Collecti
   if ( newParent.id() > 0 ) { // not root
     resourceId = newParent.resourceId();
   }
-  if ( !CollectionQueryHelper::canBeMovedTo ( collection, newParent ) ) {
+  if ( !CollectionQueryHelper::canBeMovedTo( collection, newParent ) ) {
     return false;
   }
 
@@ -563,7 +747,7 @@ bool Akonadi::DataStore::moveCollection( Collection & collection, const Collecti
   return true;
 }
 
-bool DataStore::appendMimeTypeForCollection( qint64 collectionId, const QStringList & mimeTypes )
+bool DataStore::appendMimeTypeForCollection( qint64 collectionId, const QStringList &mimeTypes )
 {
   if ( mimeTypes.isEmpty() ) {
     return true;
@@ -598,7 +782,7 @@ bool DataStore::appendMimeTypeForCollection( qint64 collectionId, const QStringL
   return true;
 }
 
-void DataStore::activeCachePolicy(Collection & col)
+void DataStore::activeCachePolicy( Collection &col )
 {
   if ( !col.cachePolicyInherit() ) {
     return;
@@ -620,16 +804,16 @@ void DataStore::activeCachePolicy(Collection & col)
   col.setCachePolicyCheckInterval( -1 );
   col.setCachePolicyCacheTimeout( -1 );
   col.setCachePolicySyncOnDemand( false );
-  col.setCachePolicyLocalParts( QLatin1String("ALL") );
+  col.setCachePolicyLocalParts( QLatin1String( "ALL" ) );
 }
 
-QVector<Collection> DataStore::virtualCollections( const PimItem& item )
+QVector<Collection> DataStore::virtualCollections( const PimItem &item )
 {
   QueryBuilder qb( CollectionPimItemRelation::tableName(), QueryBuilder::Select );
   qb.addJoin( QueryBuilder::InnerJoin, Collection::tableName(),
               Collection::idFullColumnName(), CollectionPimItemRelation::leftFullColumnName() );
   Q_FOREACH ( const QString &columnName, Collection::fullColumnNames() ) {
-    qb.addColumn( columnName  );
+    qb.addColumn( columnName );
   }
   qb.addValueCondition( CollectionPimItemRelation::rightFullColumnName(), Query::Equals, item.id() );
 
@@ -681,7 +865,7 @@ QMap<Entity::Id,PimItem> DataStore::virtualCollections( const PimItem::List &ite
 }
 
 /* --- MimeType ------------------------------------------------------ */
-bool DataStore::appendMimeType( const QString & mimetype, qint64 *insertId )
+bool DataStore::appendMimeType( const QString &mimetype, qint64 *insertId )
 {
   if ( MimeType::exists( mimetype ) ) {
     akDebug() << "Cannot insert mimetype " << mimetype
@@ -694,13 +878,13 @@ bool DataStore::appendMimeType( const QString & mimetype, qint64 *insertId )
 }
 
 /* --- PimItem ------------------------------------------------------- */
-bool DataStore::appendPimItem( QVector<Part> & parts,
-                               const MimeType & mimetype,
-                               const Collection & collection,
-                               const QDateTime & dateTime,
-                               const QString & remote_id,
-                               const QString & remoteRevision,
-                               const QString & gid,
+bool DataStore::appendPimItem( QVector<Part> &parts,
+                               const MimeType &mimetype,
+                               const Collection &collection,
+                               const QDateTime &dateTime,
+                               const QString &remote_id,
+                               const QString &remoteRevision,
+                               const QString &gid,
                                PimItem &pimItem )
 {
   pimItem.setMimeTypeId( mimetype.id() );
@@ -727,15 +911,15 @@ bool DataStore::appendPimItem( QVector<Part> & parts,
   // insert every part
   if ( !parts.isEmpty() ) {
     //don't use foreach, the caller depends on knowing the part has changed, see the Append handler
-    for (QVector<Part>::iterator it = parts.begin(); it != parts.end(); ++it ) {
+    for ( QVector<Part>::iterator it = parts.begin(); it != parts.end(); ++it ) {
 
-      (*it).setPimItemId( pimItem.id() );
-      if ( (*it).datasize() < (*it).data().size() ) {
-        (*it).setDatasize( (*it).data().size() );
+      ( *it ).setPimItemId( pimItem.id() );
+      if ( ( *it ).datasize() < ( *it ).data().size() ) {
+        ( *it ).setDatasize( ( *it ).data().size() );
       }
 
 //      akDebug() << "Insert from DataStore::appendPimItem";
-      if ( !PartHelper::insert(&(*it)) ) {
+      if ( !PartHelper::insert( &( *it ) ) ) {
         return false;
       }
     }
@@ -770,7 +954,11 @@ bool DataStore::unhideAllPimItems()
 
   akDebug() << "DataStore::unhideAllPimItems()";
 
-  return PartHelper::remove( Part::nameFullColumnName(), QLatin1String( AKONADI_ATTRIBUTE_HIDDEN ) );
+  try {
+    return PartHelper::remove( Part::partTypeIdFullColumnName(), PartTypeHelper::fromName( "ATR", "HIDDEN" ).id() );
+  } catch ( ... ) {} // we can live with this failing
+
+  return false;
 }
 
 bool DataStore::cleanupPimItems( const PimItem::List &items )
@@ -798,7 +986,7 @@ bool DataStore::cleanupPimItems( const PimItem::List &items )
   return true;
 }
 
-bool DataStore::addCollectionAttribute(const Collection & col, const QByteArray & key, const QByteArray & value)
+bool DataStore::addCollectionAttribute( const Collection &col, const QByteArray &key, const QByteArray &value )
 {
   SelectQueryBuilder<CollectionAttribute> qb;
   qb.addValueCondition( CollectionAttribute::collectionIdColumn(), Query::Equals, col.id() );
@@ -825,7 +1013,7 @@ bool DataStore::addCollectionAttribute(const Collection & col, const QByteArray
   return true;
 }
 
-bool Akonadi::DataStore::removeCollectionAttribute(const Collection & col, const QByteArray & key)
+bool DataStore::removeCollectionAttribute( const Collection &col, const QByteArray &key )
 {
   SelectQueryBuilder<CollectionAttribute> qb;
   qb.addValueCondition( CollectionAttribute::collectionIdColumn(), Query::Equals, col.id() );
@@ -841,7 +1029,7 @@ bool Akonadi::DataStore::removeCollectionAttribute(const Collection & col, const
     }
   }
 
-  if (!result.isEmpty()) {
+  if ( !result.isEmpty() ) {
     mNotificationCollector->collectionChanged( col, QList<QByteArray>() << key );
     return true;
   }
@@ -858,8 +1046,7 @@ void DataStore::debugLastDbError( const char *actionDescription ) const
                          QString::fromLatin1( "%1\nDriver said: %2\nDatabase said:%3" )
                             .arg( QString::fromLatin1( actionDescription ) )
                             .arg( m_database.lastError().driverText() )
-                            .arg( m_database.lastError().databaseText() )
-                       );
+                            .arg( m_database.lastError().databaseText() ) );
 }
 
 void DataStore::debugLastQueryError( const QSqlQuery &query, const char *actionDescription ) const
@@ -872,27 +1059,107 @@ void DataStore::debugLastQueryError( const QSqlQuery &query, const char *actionD
   Tracer::self()->error( "DataStore (Database Query Error)",
                          QString::fromLatin1( "%1: %2" )
                             .arg( QString::fromLatin1( actionDescription ) )
-                            .arg( query.lastError().text() )
-                       );
+                            .arg( query.lastError().text() ) );
 }
 
 // static
-QString DataStore::dateTimeFromQDateTime( const QDateTime & dateTime )
+QString DataStore::dateTimeFromQDateTime( const QDateTime &dateTime )
 {
     QDateTime utcDateTime = dateTime;
     if ( utcDateTime.timeSpec() != Qt::UTC ) {
         utcDateTime.toUTC();
     }
-    return utcDateTime.toString( QLatin1String("yyyy-MM-dd hh:mm:ss") );
+    return utcDateTime.toString( QLatin1String( "yyyy-MM-dd hh:mm:ss" ) );
 }
 
 // static
-QDateTime DataStore::dateTimeToQDateTime( const QByteArray & dateTime )
+QDateTime DataStore::dateTimeToQDateTime( const QByteArray &dateTime )
 {
-    return QDateTime::fromString( QString::fromLatin1(dateTime), QLatin1String("yyyy-MM-dd hh:mm:ss") );
+    return QDateTime::fromString( QString::fromLatin1( dateTime ), QLatin1String( "yyyy-MM-dd hh:mm:ss" ) );
 }
 
-bool Akonadi::DataStore::beginTransaction()
+void DataStore::addQueryToTransaction( const QSqlQuery &query, bool isBatch )
+{
+  DbType::Type dbType = DbType::type( m_database );
+  // This is used for replaying deadlocked transactions, so only record queries
+  // for backends that support concurrent transactions.
+  if ( !inTransaction() || ( dbType != DbType::MySQL && dbType != DbType::PostgreSQL ) ) {
+    return;
+  }
+
+  m_transactionQueries.append( qMakePair( query, isBatch ) );
+}
+
+QSqlQuery DataStore::retryLastTransaction()
+{
+  DbType::Type dbType = DbType::type( m_database );
+  if ( !inTransaction() || ( dbType != DbType::MySQL && dbType != DbType::PostgreSQL ) ) {
+    return QSqlQuery();
+  }
+
+  // The database has rolled back the actual transaction, so reset the counter
+  // to 0 and start a new one in beginTransaction(). Then restore the level
+  // because this has to be completely transparent to the original caller
+  const int oldTransactionLevel = m_transactionLevel;
+  m_transactionLevel = 0;
+  if ( !beginTransaction() ) {
+    m_transactionLevel = oldTransactionLevel;
+    return QSqlQuery();
+  }
+  m_transactionLevel = oldTransactionLevel;
+
+  QSqlQuery ret;
+  typedef QPair<QSqlQuery, bool> QueryBoolPair;
+  QMutableVectorIterator<QueryBoolPair> iter( m_transactionQueries );
+  while ( iter.hasNext() ) {
+    iter.next();
+    QSqlQuery query = iter.value().first;
+    const bool isBatch = iter.value().second;
+
+    // Make sure the query is ready to be executed again
+    if ( query.isActive() ) {
+      query.finish();
+    }
+
+    bool res = false;
+    if ( isBatch ) {
+      // QSqlQuery::execBatch() does not reset lastError(), so for the sake
+      // of transparency (make it look to the caller like if the query was
+      // successful the first time), we create a copy of the original query,
+      // which has lastError empty.
+      QSqlQuery copiedQuery( m_database );
+      copiedQuery.prepare( query.executedQuery() );
+      const QVariantList values = query.boundValues().values();
+      for (int i = 0; i < values.size(); ++i) {
+        copiedQuery.bindValue(i, values[i]);
+      }
+      query = copiedQuery;
+      res = query.execBatch();
+    } else {
+      res = query.exec();
+    }
+
+    if ( !res ) {
+      // Don't do another deadlock detection here, just give up.
+      akError() << "DATABASE ERROR:";
+      akError() << "  Error code:" << query.lastError().number();
+      akError() << "  DB error: " << query.lastError().databaseText();
+      akError() << "  Error text:" << query.lastError().text();
+      akError() << "  Query:" << query.executedQuery();
+
+      // Return the last query, because that's what caller expects to retrieve
+      // from QueryBuilder. It is in error state anyway.
+      return m_transactionQueries.last().first;
+    }
+
+    // Update the query in the list
+    iter.setValue( qMakePair( query, isBatch ) );
+  }
+
+  return m_transactionQueries.last().first;
+}
+
+bool DataStore::beginTransaction()
 {
   if ( !m_dbOpened ) {
     return false;
@@ -908,12 +1175,12 @@ bool Akonadi::DataStore::beginTransaction()
     }
   }
 
-  m_transactionLevel++;
+  ++m_transactionLevel;
 
   return true;
 }
 
-bool Akonadi::DataStore::rollbackTransaction()
+bool DataStore::rollbackTransaction()
 {
   if ( !m_dbOpened ) {
     return false;
@@ -924,7 +1191,7 @@ bool Akonadi::DataStore::rollbackTransaction()
     return false;
   }
 
-  m_transactionLevel--;
+  --m_transactionLevel;
 
   if ( m_transactionLevel == 0 ) {
     QSqlDriver *driver = m_database.driver();
@@ -934,14 +1201,15 @@ bool Akonadi::DataStore::rollbackTransaction()
       debugLastDbError( "DataStore::rollbackTransaction" );
       return false;
     }
-
     TRANSACTION_MUTEX_UNLOCK;
+
+    m_transactionQueries.clear();
   }
 
   return true;
 }
 
-bool Akonadi::DataStore::commitTransaction()
+bool DataStore::commitTransaction()
 {
   if ( !m_dbOpened ) {
     return false;
@@ -962,13 +1230,15 @@ bool Akonadi::DataStore::commitTransaction()
       TRANSACTION_MUTEX_UNLOCK;
       Q_EMIT transactionCommitted();
     }
+
+    m_transactionQueries.clear();
   }
 
   m_transactionLevel--;
   return true;
 }
 
-bool Akonadi::DataStore::inTransaction() const
+bool DataStore::inTransaction() const
 {
   return m_transactionLevel > 0;
 }
