@@ -34,6 +34,7 @@ class QTimer;
 #include "notificationcollector.h"
 
 namespace Akonadi {
+namespace Server {
 
 class NotificationCollector;
 
@@ -114,14 +115,20 @@ class DataStore : public QObject
     /**
       Per thread singleton.
     */
-    static DataStore* self();
+    static DataStore *self();
 
     /* --- ItemFlags ----------------------------------------------------- */
     bool setItemsFlags( const PimItem::List &items, const QVector<Flag> &flags );
-    bool appendItemsFlags( const PimItem::List &items, const QVector<Flag> &flags, bool& flagsChanged,
-                          bool checkIfExists = true, const Collection &col = Collection() );
+    bool appendItemsFlags( const PimItem::List &items, const QVector<Flag> &flags, bool &flagsChanged,
+                           bool checkIfExists = true, const Collection &col = Collection() );
     bool removeItemsFlags( const PimItem::List &items, const QVector<Flag> &flags );
 
+    /* --- ItemTags ----------------------------------------------------- */
+    bool setItemsTags( const PimItem::List &items, const Tag::List &tags );
+    bool appendItemsTags( const PimItem::List &items, const Tag::List &tags, bool &tagsChanged,
+                           bool checkIfExists = true, const Collection &col = Collection() );
+    bool removeItemsTags( const PimItem::List &items, const Tag::List &tags );
+
     /* --- ItemParts ----------------------------------------------------- */
     bool removeItemParts( const PimItem &item, const QList<QByteArray> &parts );
 
@@ -137,11 +144,11 @@ class DataStore : public QObject
     bool cleanupCollection_slow( Collection &collection );
 
     /// moves the collection @p collection to @p newParent.
-    bool moveCollection( Akonadi::Collection& collection, const Akonadi::Collection& newParent );
+    bool moveCollection( Collection &collection, const Collection &newParent );
 
-    bool appendMimeTypeForCollection( qint64 collectionId, const QStringList & mimeTypes );
+    bool appendMimeTypeForCollection( qint64 collectionId, const QStringList &mimeTypes );
 
-    static QString collectionDelimiter() { return QLatin1String("/"); }
+    static QString collectionDelimiter() { return QLatin1String( "/" ); }
 
     /**
       Determines the active cache policy for this Collection.
@@ -155,16 +162,16 @@ class DataStore : public QObject
     QMap<Entity::Id /* collection */, PimItem> virtualCollections( const PimItem::List &items );
 
     /* --- MimeType ------------------------------------------------------ */
-    bool appendMimeType( const QString & mimetype, qint64 *insertId = 0 );
+    bool appendMimeType( const QString &mimetype, qint64 *insertId = 0 );
 
     /* --- PimItem ------------------------------------------------------- */
-    bool appendPimItem( QVector<Part> & parts,
-                        const MimeType & mimetype,
-                        const Collection & collection,
-                        const QDateTime & dateTime,
-                        const QString & remote_id,
-                        const QString & remoteRevision,
-                        const QString & gid,
+    bool appendPimItem( QVector<Part> &parts,
+                        const MimeType &mimetype,
+                        const Collection &collection,
+                        const QDateTime &dateTime,
+                        const QString &remote_id,
+                        const QString &remoteRevision,
+                        const QString &gid,
                         PimItem &pimItem );
 
     /**
@@ -232,7 +239,7 @@ class DataStore : public QObject
       Returns the notification collector of this DataStore object.
       Use this to listen to change notification signals.
     */
-    NotificationCollector* notificationCollector() const { return mNotificationCollector; }
+    NotificationCollector *notificationCollector() const { return mNotificationCollector; }
 
     /**
       Returns the QSqlDatabase object. Use this for generating queries yourself.
@@ -267,20 +274,49 @@ protected:
     bool doAppendItemsFlag( const PimItem::List &items, const Flag &flag,
                            const QSet<PimItem::Id> &existing, const Collection &col );
 
+    bool doAppendItemsTag( const PimItem::List &items, const Tag &tag,
+                          const QSet<Entity::Id> &existing, const Collection &col );
+
     /** Converts the given date/time to the database format, i.e.
         "YYYY-MM-DD HH:MM:SS".
         @param dateTime the date/time in UTC
         @return the date/time in database format
         @see dateTimeToQDateTime
      */
-    static QString dateTimeFromQDateTime( const QDateTime & dateTime );
+    static QString dateTimeFromQDateTime( const QDateTime &dateTime );
 
     /** Converts the given date/time from database format to QDateTime.
         @param dateTime the date/time in database format
         @return the date/time as QDateTime
         @see dateTimeFromQDateTime
      */
-    static QDateTime dateTimeToQDateTime( const QByteArray & dateTime );
+    static QDateTime dateTimeToQDateTime( const QByteArray &dateTime );
+
+
+    /**
+     * Adds the @p query to current transaction, so that it can be replayed in
+     * case the transaction deadlocks or timeouts.
+     *
+     * When DataStore is not in transaction or SQLite is configured, this method
+     * does nothing.
+     *
+     * All queries will automatically be removed when transaction is committed.
+     *
+     * This method should only be used by QueryBuilder.
+     */
+    void addQueryToTransaction( const QSqlQuery &query, bool isBatch );
+
+    /**
+     * Tries to execute all queries from last transaction again. If any of the
+     * queries fails, the entire transaction is rolled back and fails.
+     *
+     * This method can only be used by QueryBuilder when database rolls back
+     * transaction due to deadlock or timeout.
+     *
+     * @return Returns an invalid query when error occurs, or the last replayed
+     *         query on success.
+     */
+    QSqlQuery retryLastTransaction();
 
   private Q_SLOTS:
     void sendKeepAliveQuery();
@@ -290,11 +326,18 @@ private:
     QSqlDatabase m_database;
     bool m_dbOpened;
     uint m_transactionLevel;
+    QVector<QPair<QSqlQuery,bool /* isBatch */> > m_transactionQueries;
     QByteArray mSessionId;
-    NotificationCollector* mNotificationCollector;
+    NotificationCollector *mNotificationCollector;
     QTimer *m_keepAliveTimer;
     static bool s_hasForeignKeyConstraints;
+
+    // Gives QueryBuilder access to addQueryToTransaction() and retryLastTransaction()
+    friend class QueryBuilder;
+
 };
 
-}
+} // namespace Server
+} // namespace Akonadi
+
 #endif
